# This file is part of hive, a distributed bug tracker
# Copyright (c) 2019 by Adam Hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Functions for working with hive's data format (getting information about an
issue, etc).
"""

import os
import ast
import sys
import json
import time
import pickle
import hashlib
import contextlib

import tqdm

from uuid import uuid4
from datetime import datetime

DATA_VERSION = 1


class HiveDataException(Exception):
    pass


def json_dump(x, f):
    return json.dump(
        x,
        f,
        sort_keys=True,
        indent=2,
        default=lambda x: list(x) if isinstance(x, set) else None,
    )


def json_format(x):
    return json.dumps(
        x,
        sort_keys=True,
        indent=2,
        default=lambda x: list(x) if isinstance(x, set) else None,
    )


def initialize(repo_root):
    """
    set up hive's initial structure in repo_root/.hive-bt
    """
    hive_root = os.path.join(repo_root, ".hive-bt")
    if os.path.exists(hive_root):
        raise HiveDataException("hive repository already exists at %s" % hive_root)

    os.makedirs(hive_root, exist_ok=True)
    with open(os.path.join(hive_root, "version"), "w") as f:
        f.write("hive-%d\n" % DATA_VERSION)

    for subdir in ("issues", "labels", "targets"):
        os.makedirs(os.path.join(hive_root, subdir))
        with open(os.path.join(hive_root, subdir, ".keep"), "w") as f:
            pass


def get_issue_location(repo_root, issue_id):
    hive_root = os.path.join(repo_root, ".hive-bt")
    if not os.path.isdir(hive_root):
        raise HiveDataException("No hive repository here.")
    issue_id = issue_id.lower()
    possible_issues = [i for i in list_issues(repo_root) if i.startswith(issue_id)]
    if not possible_issues:
        raise HiveDataException("No bug with ID %s" % issue_id)
    elif len(possible_issues) > 2:
        raise HiveDataException(
            "Ambiguous issue ID %r (possibilities: %r)" % (issue_id, possible_issues)
        )
    issue_id = possible_issues[0]
    return issue_id, os.path.join(hive_root, "issues", issue_id)


def real_comment_id(repo_root, issue_id, comment_prefix):
    all_ids = set(get_issue_state(repo_root, issue_id)["comments_by_id"])
    possible = [i for i in all_ids if i.startswith(comment_prefix)]
    if not possible:
        raise HiveDataException("No comment with ID %s" % comment_prefix)
    elif len(possible) > 2:
        raise HiveDataException(
            "Ambiguous comment ID %r (possibilities: %r)" % (comment_prefix, possible)
        )
    return possible[0]


def list_issues(repo_root):
    hive_root = os.path.join(repo_root, ".hive-bt", "issues")
    if not os.path.isdir(hive_root):
        raise HiveDataException("No hive repository here.")
    return set(os.listdir(hive_root)) - {".keep"}


def new_issue_id(suffix):
    return "%s%s" % (uuid4().hex, suffix)


def issue_exists(repo_root, issue_id):
    return os.path.isdir(os.path.join(repo_root, ".hive-bt", "issues", issue_id))


def create_issue(repo_root, issue_id):
    try:
        os.makedirs(os.path.join(repo_root, ".hive-bt", "issues", issue_id))
    except Exception as e:
        raise HiveDataException("Could not create issue %r" % issue_id) from e


def next_logical_time(repo_root):
    hive_root = os.path.join(repo_root, ".hive-bt")
    if not os.path.isdir(hive_root):
        raise HiveDataException("Not within a hive repository.")
    issues_root = os.path.join(hive_root, "issues")
    best = -1
    for root, dirs, files in os.walk(issues_root):
        files = [f for f in files if not f.startswith(".")]
        best = max(max(int(f.split("-")[0]) for f in files) if files else -1, best)
    return best + 1


def valid_label_name(x, type_="label"):
    invalid_chars = r"\/ :."
    if not (x and all(i not in x for i in invalid_chars)):
        raise HiveDataException("invalid %s name: %r" % (type_, x))


def edit_label(repo_root, label_name, description):
    valid_label_name(label_name)
    with open(os.path.join(repo_root, ".hive-bt", "labels", label_name), "w") as f:
        json_dump(description, f)
    return label_name


def label_information(repo_root):
    label_root = os.path.join(repo_root, ".hive-bt", "labels")
    out = {}
    for i in os.listdir(label_root):
        try:
            valid_label_name(i)
        except HiveDataException:
            continue
        with open(os.path.join(label_root, i), "r") as f:
            out[i] = json.load(f)
    return out


def edit_target(repo_root, target_name, description):
    valid_target_name(target_name, type_="target")
    with open(os.path.join(repo_root, ".hive-bt", "targets", target_name), "w") as f:
        json_dump(description, f)
    return target_name


def add_action(repo_root, issue_id, action, exact=False):
    if exact:
        issue_dir = os.path.join(repo_root, ".hive-bt", "issues", issue_id)
    else:
        issue_id, issue_dir = get_issue_location(repo_root, issue_id)
    action_id = "%d-%d-%s" % (
        next_logical_time(repo_root),
        time.time() * 1000,
        hashlib.md5(json_format(action).encode("utf-8")).hexdigest(),
    )
    action["id"] = action_id
    with open(os.path.join(issue_dir, action_id), "w") as f:
        json_dump([action], f)
    return action_id


def _time_from_id(i):
    pieces = i.split("-")
    return int(pieces[0]), int(pieces[1])


def pack(repo_root, issue_id, files):
    if not files:
        raise HiveDataException("no action files specified")
    issue_root = os.path.join(repo_root, ".hive-bt", "issues", issue_id)
    if files == "all":
        files = os.listdir(issue_root)
    if len(files) == 1:
        raise HiveDataException("only one file specified; already packed!")
    files = [os.path.join(issue_root, fname) for fname in files]
    for f in files:
        if not os.path.isfile(f):
            raise HiveDataException("no such action file: %s" % f)
    actions = []
    for fname in files:
        with open(os.path.join(issue_root, fname)) as f:
            actions.extend(json.load(f))
    if len(actions) <= 1:
        raise HiveDataException("only one action specified; not packing")
    actions.sort(key=lambda x: _time_from_id(x["id"]))
    pack_time = _time_from_id(actions[-1]["id"])
    pack_contents = json_format(actions)
    hash_ = hashlib.md5(pack_contents.encode("utf-8")).hexdigest()
    pack_fname = "%d-%d-%s" % (*pack_time, hash_)
    with open(os.path.join(issue_root, pack_fname), "w") as f:
        f.write(pack_contents)
    for fname in files:
        if fname != pack_fname:
            os.unlink(os.path.join(issue_root, fname))
    return pack_fname, len(actions)


def unpack(repo_root, issue_id, fname):
    issue_root = os.path.join(repo_root, ".hive-bt", "issues", issue_id)
    fname = os.path.join(issue_root, fname)
    with open(fname) as f:
        actions = json.load(f)
    if len(actions) <= 1:
        raise HiveDataException("only one action specified; nothing to unpack")
    ids = []
    for action in actions:
        with open(os.path.join(issue_root, action["id"]), "w") as f:
            json_dump([action], f)
        ids.append(action["id"])
    os.unlink(fname)
    return ids, len(actions)


def get_cache_location():
    home_dir = os.path.expanduser("~")
    if "XDG_CACHE_HOME" in os.environ:
        return os.environ["XDG_CACHE_HOME"]
    elif sys.platform.startswith("darwin"):
        return os.path.join(home_dir, "Library", "Caches", "hive_bt")
    elif sys.platform.startswith("win32"):
        return os.path.join(os.environ["LOCALAPPDATA"], "hive_bt")
    else:
        return os.path.join(home_dir, ".cache", "hive_bt")


def _splitpath(path, sofar=[]):
    folder, path = os.path.split(path)
    if path == "":
        return sofar[::-1]
    elif folder == "":
        return (sofar + [path])[::-1]
    else:
        return _splitpath(folder, sofar + [path])


def get_issue_state(repo_root, issue_id, time=None, exact=False):
    if exact:
        issue_dir = os.path.join(repo_root, ".hive-bt", "issues", issue_id)
    else:
        orig_id = issue_id
        issue_id, issue_dir = get_issue_location(repo_root, issue_id)
        if not issue_id:
            raise HiveDataException("No such issue: %r" % orig_id)
    info = {
        "id": issue_id,
        "updated_times": [],
        "labels": set(),
        "assignees": set(),
        "priority": 0,
        "target": None,
        "open": True,
        "comments": [],
    }
    comments_by_id = {}
    info["comments_by_id"] = comments_by_id
    actions = []
    for filename in os.listdir(issue_dir):
        with open(os.path.join(issue_dir, filename)) as f:
            this_pack_actions = json.load(f)
        for action in this_pack_actions:
            pieces = action["id"].split("-")
            actions.append(
                (int(pieces[0]), int(pieces[1]), pieces[2], action["id"], action)
            )
    actions.sort()
    for logicaltime, unixtime, hash_, action_id, action in actions:
        if time is not None and logicaltime > time:
            break
        unixtime /= 1000
        action_name = action["action"]
        info["labels"] = set(info["labels"])

        # decide how to update the state based on the given action
        if action_name == "set_parameters":
            info.update(action["parameters"])
            info["labels"] = set(info["labels"])

        elif action_name == "add_labels":
            info["labels"] = set(info["labels"]) | set(action["labels"])

        elif action_name == "remove_labels":
            info["labels"] = set(info["labels"]) - set(action["labels"])

        elif action_name == "add_assignees":
            info["assignees"] = set(info["assignees"]) | set(action["assignees"])

        elif action_name == "remove_assignees":
            info["assignees"] = set(info["assignees"]) - set(action["assignees"])

        elif action_name == "open":
            info["open"] = True
        elif action_name == "close":
            info["open"] = False

        elif action_name == "add_comment":
            comment = comments_by_id[action_id] = {
                k: v
                for k, v in action.items()
                if k not in {"action", "parent", "author"}
            }
            comment["authors"] = [action["author"]]
            comment["edited"] = False
            comment["id"] = action_id
            comment["display_time"] = unixtime

            parent = action.get("parent", None)
            if parent in comments_by_id:
                comments_by_id[parent].setdefault("children", []).append(comment)
            else:
                info["comments"].append(comment)
        elif action_name == "edit_comment":
            comments_by_id[action["id"]].update(
                {
                    k: v
                    for k, v in action.items()
                    if k not in {"action", "parent", "author"}
                }
            )
            comments_by_id[action["id"]]["authors"].append(action["author"])
            comments_by_id[action["id"]]["edited"] = True

        # update timestamps
        info["updated_times"].append((logicaltime, unixtime))

    return info


sort_funcs = {
    "priority": (lambda x: x.get("priority", 0) or 0, True),
    "created_time": (lambda x: x["updated_times"][0], True),
    "updated_time": (lambda x: x["updated_times"][-1], True),
    "title": (lambda x: x["title"].lower(), False),
    "status": (lambda x: 1 if x["open"] else -1, True),
}


def _any_filter(i, q):
    return any(
        filter_funcs[filt](i, q) for filt in ("body", "title", "author", "label")
    )


filter_funcs = {
    "priority>": lambda i, q: sort_funcs["priority"](i) > ast.literal_eval(q),
    "priority<": lambda i, q: sort_funcs["priority"](i) < ast.literal_eval(q),
    "priority": lambda i, q: sort_funcs["priority"](i) == ast.literal_eval(q),
    "label": lambda i, q: q in i["labels"],
    "status": lambda i, q: True
    if q == "any"
    else i["open"]
    if q == "open"
    else not i["open"]
    if q == "closed"
    else False,
    "body": lambda i, q: q.lower() in i["body"].lower(),
    "title": lambda i, q: q.lower() in i["title"].lower(),
    "author": lambda i, q: q.lower() in i["author"].lower(),
    "id": lambda i, q: i["id"].lower().startswith(q.lower()),
    "any": _any_filter,
}


@contextlib.contextmanager
def atomic_write(fname, mode="w"):
    new_name = "%s.%s.part" % (fname, uuid4().hex)
    with open(new_name, mode) as f:
        yield f
        f.flush()
    os.replace(new_name, fname)


def cached_all_issues(repo_root, override_cache=False, progress=False):
    cache_dir = os.path.join(get_cache_location(), *_splitpath(repo_root), ".hive-bt")
    cache_loc = os.path.join(cache_dir, "issues.pickle")

    max_mtime = -float("inf")
    for root, dirs, files in os.walk(os.path.join(repo_root, ".hive-bt", "issues")):
        max_mtime = max(
            max_mtime,
            os.stat(root).st_mtime,
            max(os.stat(os.path.join(root, f)).st_mtime for f in files)
            if files
            else -1,
        )

    if (
        (not override_cache)
        and os.path.isfile(cache_loc)
        and os.stat(cache_loc).st_mtime > max_mtime
    ):
        try:
            with open(cache_loc, "rb") as f:
                return pickle.load(f)
        except:
            pass
    all_issues = []
    list_ = list_issues(repo_root)
    if progress:
        list_ = tqdm.tqdm(list_)
    for issue in list_:
        all_issues.append(get_issue_state(repo_root, issue, exact=True))
    os.makedirs(cache_dir, exist_ok=True)
    with atomic_write(cache_loc, "wb") as f:
        pickle.dump(all_issues, f, pickle.HIGHEST_PROTOCOL)
    return all_issues


def all_issues_state(
    repo_root, sorts=None, reverse_sort=False, filters=None, include_closed=False
):
    all_issues = cached_all_issues(repo_root)
    filters = list(filters or [])
    if not include_closed and "status" not in {i[0] for i in filters}:
        filters.append(("status", "open"))
    sorts = sorts or ["priority", "created_time"]
    for func, query in filters:
        try:
            all_issues = [
                issue for issue in all_issues if filter_funcs[func](issue, query)
            ]
        except KeyError:
            raise HiveDataException("No filter function called %r" % func)
    for func in reversed(sorts):
        try:
            func, reverse = sort_funcs[func]
            if reverse_sort:
                reverse = not reverse
            all_issues.sort(key=func, reverse=reverse)
        except KeyError:
            raise HiveDataException("No sorting function called %r" % func)
    return all_issues



# This file is part of hive, a distributed bug tracker
# Copyright (c) 2019 by Adam Hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Helper utilities for working with version control systems.  Currently works by
invoking the binaries directly.
"""

import os
import shutil
import subprocess

from . import data


def has_git():
    return shutil.which("git") is not None


def has_hg():
    return shutil.which("hg") is not None


def _subproc_output(cmd):
    try:
        return subprocess.check_output(cmd, stderr=subprocess.PIPE).strip().decode()
    except subprocess.CalledProcessError:
        return None


def hg_root():
    if not has_hg():
        return None
    return _subproc_output(["hg", "root"])


def git_root():
    if not has_git():
        return None
    return _subproc_output(["git", "rev-parse", "--show-toplevel"])


root_funcs = {"git": git_root, "hg": hg_root}


def get_root():
    for type_, func in root_funcs.items():
        root = func()
        if root:
            return (type_, root)
    return None


def hg_name():
    if not has_hg():
        return None
    return _subproc_output(["hg", "config", "ui.username"])


def git_name():
    if not has_git():
        return None
    name = _subproc_output(["git", "config", "--get", "user.name"])
    email = _subproc_output(["git", "config", "--get", "user.email"])
    if name is None and email is None:
        return None
    elif name is None:
        return email
    elif email is None:
        return name
    return "%s <%s>" % (name, email)


name_funcs = {"hg": hg_name, "git": git_name}


def get_repo_info():
    repo = get_root()
    if not repo:
        return None
    ident = name_funcs[repo[0]]()
    return {"type": repo[0], "root": repo[1], "name": ident}


def modified_issues(repo_info):
    hive_root = os.path.join(repo_info["root"], ".hive-bt", "issues")
    type_ = repo_info["type"]
    if type_ == "hg":
        files = (
            subprocess.check_output(["hg", "status", "."], cwd=hive_root)
            .decode()
            .splitlines()
        )
    elif type_ == "git":
        files = (
            subprocess.check_output(
                ["git", "status", "--porcelain=2", "--short", "."], cwd=hive_root
            )
            .decode()
            .splitlines()
        )
    return set(i.strip().split(" ", 1)[1].split(os.sep, 1)[0] for i in files) - {
        ".keep"
    }


def untracked_actions(repo_info):
    hive_root = os.path.join(repo_info["root"], ".hive-bt", "issues")
    type_ = repo_info["type"]
    if type_ == "hg":
        files = (
            subprocess.check_output(["hg", "status", "-u", "-n", "."], cwd=hive_root)
            .decode()
            .splitlines()
        )
    elif type_ == "git":
        files = (
            subprocess.check_output(
                ["git", "status", "--porcelain=2", "-uall", "."], cwd=hive_root
            )
            .decode()
            .splitlines()
        )
        files = [i[1] for i in [j.split(" ", 1) for j in files] if i[0] == "?"]
    action_dict = {}
    for f in files:
        if f == ".keep":
            continue
        issue, action = f.split(os.sep, 1)
        action_dict.setdefault(issue, []).append(action)
    return action_dict


def pack(repo_info, untracked_only):
    if untracked_only:
        to_pack = untracked_actions(repo_info)
    else:
        to_pack = {k: "all" for k in data.list_issues(repo_info["root"])}

    for issue, actions in to_pack.items():
        try:
            new_id, num_actions = data.pack(repo_info["root"], issue, actions)
        except data.HiveDataException:
            continue
        print(
            "Combined %d actions in issue %s into a single file" % (num_actions, issue)
        )


def add(repo_info):
    hive_root = os.path.join(repo_info["root"], ".hive-bt")
    type_ = repo_info["type"]
    if type_ == "hg":
        command = ["hg", "addremove", "."]
    elif type_ == "git":
        command = ["git", "add", "."]
    subprocess.run(command, cwd=hive_root)


def commit(repo_info, msg=None):
    if msg is None:
        msg = "Update issues in .hive-bt"
    hive_root = os.path.join(repo_info["root"], ".hive-bt")
    subprocess.run([repo_info["type"], "commit", "-m", msg, "."], cwd=hive_root)



# This file is part of hive, a distributed bug tracker
# Copyright (c) 2019 by Adam Hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Command-line Interface for hive
"""

import io
import os
import re
import sys
import click
import shutil
import textwrap
import termcolor
import webbrowser

from datetime import datetime

from . import vcs
from . import util
from . import data
from . import commands
from . import __version__

from .web import app


def colored(text, fgcolor=None, bgcolor=None, **kwargs):
    if not sys.stdout.isatty():
        return text
    elif fgcolor is not None and bgcolor is not None:
        return "\x1b[38;5;%d;48;5;%dm%s\x1b[0m" % (
            util.color_256(fgcolor),
            util.color_256(bgcolor),
            text,
        )
    return termcolor.colored(text, fgcolor, **kwargs)


def body_from_stdin_or_editor(stdin, initial_md="", quoted_md="", allow_empty=False):
    if stdin:
        body = sys.stdin.read()
    else:
        body = util.get_editor_markdown(initial_md=initial_md, quoted_md=quoted_md)
        body = body.split("--", 1)[0].strip()
        if (not body) and (not allow_empty):
            click.echo(colored("No body text supplied.", "red"), err=True)
            sys.exit(1)
        elif body == initial_md.strip():
            click.echo(colored("No change to body text.", "red"), err=True)
            sys.exit(1)
    return body


_alias_map = {"ls": "list", "tag": "label"}

# from http://click.palletsprojects.com/en/5.x/advanced/
class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        if cmd_name in _alias_map:
            cmd_name = _alias_map[cmd_name]
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))


@click.command(cls=AliasedGroup)
def main():
    pass


def is_repo_helper(repo_root):
    hive_loc = os.path.join(repo_root, ".hive-bt")
    if not os.path.isdir(hive_loc):
        click.echo(
            colored(
                'ERROR: no hive repository at %s; run "hive init" to initialize it'
                % (hive_loc),
                "red",
            ),
            err=True,
        )
        sys.exit(1)


def find_repo_helper(error_on_no_identity=False, error_on_no_hive=True):
    repo_info = vcs.get_repo_info()
    if repo_info is None:
        click.echo(
            colored(
                "ERROR: hive must be run from within a Git or Mercurial repository",
                "red",
            ),
            err=True,
        )
        sys.exit(1)
    if error_on_no_hive and not os.path.isdir(
        os.path.join(repo_info["root"], ".hive-bt", "issues")
    ):
        click.echo(colored("ERROR: no hive issue repository here.", "red"))
        sys.exit(1)
    if repo_info["name"] is None:
        color = "red" if error_on_no_identity else "magenta"
        click.echo(
            colored("ERROR: identity not set in gitconfig or hgrc", color), err=True
        )
        if error_on_no_identity:
            sys.exit(1)
    return repo_info


@click.command()
def init():
    """
    Initialize a hive repository
    """
    repo_info = find_repo_helper(error_on_no_hive=False)
    try:
        commands.initialize(repo_info["root"])
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)
    click.echo(
        colored(
            "Initialized hive repository at %s"
            % os.path.join(repo_info["root"], ".hive-bt"),
            "green",
        )
    )


def uline(x):
    return colored(x, attrs=["underline"])


def print_comment(c, indent="  | ", limit=80):
    print(indent + uline("Comment ID:"), c["id"])
    print(indent + uline("Author:"), c["authors"][0])
    print(indent + uline("Created:"), util.format_time(c["display_time"]))
    print(indent)
    print(
        textwrap.TextWrapper(
            break_long_words=False,
            initial_indent=indent,
            subsequent_indent=indent,
            width=limit,
        ).fill(c["body"])
    )
    print(indent)
    for child in c.get("children", []):
        print(indent)
        print(indent)
        print_comment(child, indent="%s  | " % indent, limit=limit)


@click.command()
@click.argument("issue")
def show(issue):
    """
    Display information about a particular issue
    """
    repo_info = find_repo_helper()
    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)
    print(uline("Issue ID:"), issue["id"])
    print(uline("Title:"), issue["title"].strip())
    print(uline("Submitter:"), issue["author"])
    if issue.get("reporter", None) is not None:
        print(uline("Reporter:"), issue["reporter"])
    print(uline("Created:"), util.format_time(issue["updated_times"][0][1]))
    if len(issue["updated_times"]) > 1:
        print(uline("Last Updated:"), util.format_time(issue["updated_times"][-1][1]))
    print(uline("Priority:"), issue["priority"])
    print(
        uline("Labels:"),
        format_labels(
            sorted(issue["labels"]), data.label_information(repo_info["root"])
        ),
    )
    print(uline("Assignees:"), ", ".join(issue["assignees"]))
    print()
    print(
        textwrap.TextWrapper(
            break_long_words=False, width=shutil.get_terminal_size()[0]
        ).fill(issue["body"])
    )
    for comment in issue["comments"]:
        print()
        print()
        print_comment(comment, limit=shutil.get_terminal_size()[0])


@click.command(name="list")
@click.argument("search_term", nargs=-1)
@click.option(
    "-s",
    "--sort-by",
    multiple=True,
    help="Function to sort by.  Specifying multiple times will set secondary, tertiary, ..., sorts..",
    type=click.Choice(data.sort_funcs),
)
@click.option(
    "--long-ids",
    is_flag=True,
    help="Show full issue IDs, rather than minimally short IDs",
)
@click.option("-r", "--reverse-sort", is_flag=True, help="Sort in reverse order.")
@click.option(
    "--max-title-length", type=int, help="Maximum title length to show", default=50
)
@click.option("-a", "--all", is_flag=True, help="List all issues, not just open ones")
def list_(search_term, sort_by, long_ids, reverse_sort, max_title_length, all):
    """
    Show a summary of issues
    """
    repo_info = find_repo_helper()

    filters = []
    for i in search_term:
        i = i.lower()
        for filt in data.filter_funcs:
            if i.startswith("%s:" % filt):
                filters.append(tuple(i.split(":", 1)))
                break
        else:
            filters.append(("any", i))
    try:
        all_issues = data.all_issues_state(
            repo_info["root"],
            sorts=sort_by,
            filters=filters,
            reverse_sort=reverse_sort,
            include_closed=all,
        )
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    label_info = data.label_information(repo_info["root"])
    modified_issues = vcs.modified_issues(repo_info)

    if not all_issues:
        click.echo(colored("No matching issues", "red"))
        sys.exit(0)

    longest_title = 0
    longest_priority = 0
    longest_comments = 0
    for issue in all_issues:
        if len(issue["title"]) > max_title_length - 3:
            issue["title"] = "%s..." % (issue["title"][: max_title_length - 3],)
        issue["priority"] = "P%s" % issue["priority"]
        issue["comment_str"] = "%dc" % len(issue["comments_by_id"])
        issue["oid"] = issue["id"]
        longest_title = max(len(issue["title"]), longest_title)
        longest_priority = max(len(issue["priority"]), longest_priority)
        longest_comments = max(len(issue["comment_str"]), longest_comments)

    if not long_ids:
        prefixes = util.unique_prefixes(data.list_issues(repo_info["root"]))
        for issue in all_issues:
            issue["id"] = prefixes[issue["id"]]

    max_id_length = max(len(i["id"]) for i in all_issues)

    for issue in all_issues:
        print(
            colored(
                ("{0: >.%ds}" % max_id_length).format(
                    issue["id"].rjust(max_id_length, " ")
                ),
                "green" if issue["open"] else "red",
                attrs=["bold"],
            ),
            end="  ",
        )
        print(
            ("{0: <.%ds}" % longest_title).format(
                issue["title"].ljust(longest_title, " ")
            ),
            end="  ",
        )
        print(
            datetime.fromtimestamp(issue["updated_times"][0][1]).strftime(
                "%Y-%m-%d %H:%M"
            ),
            end="  ",
        )
        print(
            ("{0: <.%ds}" % longest_priority).format(
                issue["priority"].ljust(longest_priority, " ")
            ),
            end="  ",
        )
        print(
            ("{0: <.%ds}" % longest_comments).format(
                issue["comment_str"].rjust(longest_comments, " ")
            ),
            end="  ",
        )
        print(colored("M", "red") if issue["oid"] in modified_issues else " ", end="  ")
        print(format_labels(sorted(issue["labels"]), label_info))


def format_labels(labels, label_info):
    out = []
    for l in labels:
        if l not in label_info:
            out.append((l, None, None))
        elif not all(i in label_info[l] for i in ("bgcolor", "fgcolor")):
            out.append((l, None, None))
        else:
            out.append((l, label_info[l]["fgcolor"], label_info[l]["bgcolor"]))
    return " ".join(colored(*i, attrs=["reverse"]) for i in out)


def check_priority(ctx, param, value):
    try:
        return int(value)
    except:
        try:
            return float(value)
        except:
            raise click.BadParameter("priority must be an integer or a decimal number")


@click.command(name="new")
@click.option(
    "-t",
    "--title",
    prompt="Title",
    help="the title of this issue (if none is provided, you will be prompted to enter one)",
)
@click.option(
    "-p",
    "--priority",
    callback=check_priority,
    default=0,
    help="this issue's priority, as an integer or decimal number (default: 0)",
)
@click.option(
    "-a",
    "--assignee",
    multiple=True,
    help="a person assigned to this issue (can be given multiple times)",
)
@click.option(
    "-l",
    "--label",
    multiple=True,
    help="a label associated with this issue (can be given multiple times)",
)
@click.option(
    "-r", "--target", type=str, default=None, help="this issue's target (default: None)"
)
@click.option(
    "-e",
    "--empty",
    is_flag=True,
    default=False,
    help="create an issue with an empty body",
)
@click.option(
    "-s",
    "--stdin",
    is_flag=True,
    default=False,
    help="read body from stdin rather than opening an editor",
)
def new_issue(title, priority, assignee, label, target, empty, stdin):
    """
    Create a new issue
    """
    title = title.strip()
    if not title:
        click.echo(colored("ERROR: no title supplied.", "red"), err=True)
        sys.exit(1)

    repo_info = find_repo_helper(error_on_no_identity=True)

    params = {
        "priority": priority,
        "assignees": set(assignee),
        "labels": set(label),
        "target": target,
    }

    body = "" if empty else body_from_stdin_or_editor(stdin)

    try:
        issue_id, action_id = commands.new_issue(
            repo_info["root"], repo_info["name"], title, body, params
        )
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)
    click.echo(colored("Created issue %s" % issue_id, "green"))


@click.command()
@click.argument("issue")
@click.argument("label", nargs=-1)
@click.option(
    "-d", "--delete", is_flag=True, help="Remove labels instead of adding them"
)
def label(issue, label, delete):
    """
    Add or remove labels from an issue
    """

    if not label:
        click.echo(colored("No labels provided.  Exiting"))
        sys.exit(0)

    repo_info = find_repo_helper(error_on_no_identity=True)

    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    func = commands.remove_labels if delete else commands.add_labels
    try:
        new_id = func(repo_info["root"], repo_info["name"], issue["id"], set(label))
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    if delete:
        click.echo(
            colored(
                "Removed the following labels from issue %s:" % (issue["id"],), "green"
            )
        )
    else:
        click.echo(
            colored("Added the following labels to issue %s:" % (issue["id"],), "green")
        )
    click.echo(format_labels(label, data.label_information(repo_info["root"])))


@click.command()
@click.argument("issue")
@click.argument("assignee", nargs=-1)
@click.option(
    "-d", "--delete", is_flag=True, help="Remove assignees instead of adding them"
)
@click.option(
    "-m",
    "--me",
    is_flag=True,
    help="Add/remove yourself to/from the issue, rather than providing names",
)
def assign(issue, assignee, delete, me):
    """
    Add or remove assignees from an issue
    """
    repo_info = find_repo_helper(error_on_no_identity=True)

    if me:
        assignee = [repo_info["name"]]
    if not assignee:
        click.echo(colored("No assignees provided.  Exiting"))
        sys.exit(0)

    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    func = commands.remove_assignees if delete else commands.add_assignees
    try:
        new_id = func(repo_info["root"], repo_info["name"], issue["id"], set(assignee))
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    if delete:
        click.echo(colored("Unassigned from issue %s:" % (issue["id"],), "green"))
    else:
        click.echo(colored("Assigned to issue %s:" % (issue["id"],), "green"))
    for a in assignee:
        click.echo(a)


@click.command()
@click.argument("issue")
@click.option(
    "-s",
    "--stdin",
    is_flag=True,
    default=False,
    help="read body from stdin rather than opening an editor",
)
@click.option(
    "-r",
    "--reply-to",
    type=str,
    default=None,
    help="the ID of a comment to respond to (if not specified, make a new top-level comment)",
)
@click.option(
    "-e",
    "--edit",
    type=str,
    default=None,
    help="the ID of a comment to edit (instead of making a new comment)",
)
def comment(issue, stdin, reply_to, edit):
    """
    Create or edit comments on the given issue
    """
    repo_info = find_repo_helper(error_on_no_identity=True)

    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    if edit is None:
        initial_md = ""
        allow_empty = False
    else:
        try:
            edit = data.real_comment_id(repo_info["root"], issue["id"], edit)
            initial_md = issue["comments_by_id"][edit]["body"]
            allow_empty = True
        except data.HiveDataException as e:
            click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
            sys.exit(1)

    if reply_to is not None:
        try:
            reply_to = data.real_comment_id(repo_info["root"], issue["id"], reply_to)
        except data.HiveDataException as e:
            click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
            sys.exit(1)

    body = body_from_stdin_or_editor(
        stdin, initial_md=initial_md, allow_empty=allow_empty
    )

    if edit is None:
        try:
            new_id = commands.add_comment(
                repo_info["root"], repo_info["name"], issue["id"], body, parent=reply_to
            )
        except data.HiveDataException as e:
            click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
            sys.exit(1)

        click.echo(
            colored("Created comment %s on issue %s" % (new_id, issue["id"]), "green")
        )
    else:
        try:
            new_id = commands.edit_comment(
                repo_info["root"], repo_info["name"], issue["id"], edit, body
            )
        except data.HiveDataException as e:
            click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
            sys.exit(1)

        click.echo(
            colored("Edited comment %s on issue %s" % (edit, issue["id"]), "green")
        )


def command_helper(issue, func, success_msg, *args, **kwargs):
    repo_info = find_repo_helper(error_on_no_identity=True)

    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    try:
        func(repo_info["root"], repo_info["name"], issue["id"], *args, **kwargs)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)
    click.echo(colored(success_msg(issue), "green"))


@click.command()
@click.argument("issue")
@click.argument("priority", callback=check_priority)
def priority(issue, priority):
    """
    Set an issue's priority
    """
    command_helper(
        issue,
        commands.set_priority,
        lambda i: "Set issue %s priority to %s" % (i["id"], priority),
        priority,
    )


@click.command()
@click.argument("issue")
@click.option(
    "-c",
    "--comment",
    type=str,
    default=None,
    help="A comment to make before closing the issue",
)
def close(issue, comment):
    """
    Close a given issue
    """
    repo_info = find_repo_helper(error_on_no_identity=True)

    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    if comment is not None:
        try:
            new_id = commands.add_comment(
                repo_info["root"], repo_info["name"], issue["id"], comment
            )
        except data.HiveDataException as e:
            click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
            sys.exit(1)

        click.echo(
            colored("Created comment %s on issue %s" % (new_id, issue["id"]), "green")
        )

    try:
        commands.close_issue(repo_info["root"], repo_info["name"], issue["id"])
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)
    click.echo(colored("Closed issue %s" % issue["id"], "green"))


@click.command(name="open")
@click.argument("issue")
def open_(issue):
    """
    Open/reopen a given issue
    """
    command_helper(issue, commands.open_issue, lambda i: "Opened issue %s" % (i["id"],))


@click.command()
@click.argument("issue")
@click.argument("target")
def target(issue, target):
    """
    Set an issue's target
    """
    command_helper(
        issue,
        commands.set_target,
        lambda i: "Set issue target for issue %s to %s" % (i["id"], target),
        target,
    )


@click.command()
@click.option(
    "-b",
    "--browser",
    is_flag=True,
    help="Launch a browser after starting the application",
)
@click.option(
    "--read-only",
    help="Launch the web interface in read-only mode (cannot modify issues)",
)
@click.option("-u", "--url-root", help="Root URL to prepend to links", default="")
def web(browser, read_only, url_root):
    """
    Start the web interface
    """
    repo_info = vcs.get_repo_info()
    app.config["repo_info"] = repo_info
    app.config["read_only"] = read_only
    app.config["url_root"] = url_root
    if browser:
        webbrowser.open_new("http://127.0.0.1:5000")
    app.run(debug=True)


def valid_color(ctx, param, value):
    hexchars = set("0123456789abcdef")
    value = value.strip().lower().lstrip("#")
    if value and (len(value) != 6 or any(i not in hexchars for i in value)):
        raise click.BadParameter(
            "color must be specified as 6 hex characters, e.g. ffffff"
        )
    return value


@click.command(name="edit-label")
@click.argument("label")
@click.option(
    "-d", "--description", help="description", prompt="Description", default=""
)
@click.option(
    "-b",
    "--bgcolor",
    help="background color (hex)",
    prompt="Background Color (hex, leave empty to use default)",
    callback=valid_color,
    default="",
)
@click.option(
    "-f",
    "--fgcolor",
    help="foreground color (hex)",
    prompt="Foreground (text) Color (hex, leave empty to use default)",
    callback=valid_color,
    default="",
)
def edit_label(label, description, bgcolor, fgcolor):
    """
    Edit the given label
    """
    repo_info = find_repo_helper(error_on_no_identity=True)
    vals = {"description": description, "bgcolor": bgcolor, "fgcolor": fgcolor}
    try:
        commands.edit_label(
            repo_info["root"], label, {k: v for k, v in vals.items() if v}
        )
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)
    click.echo(colored("Updated label %s" % label, "green"))


def valid_date(ctx, param, value):
    if value and not (re.match(r"\d{4}-\d{2}-\d{2}", value)):
        raise click.BadParameter("date must be specified as YYYY-MM-DD")
    return value


@click.command(name="edit-target")
@click.argument("target")
@click.option(
    "-d", "--description", help="description", prompt="Description", default=""
)
@click.option(
    "-u",
    "--due-date",
    help="due date (YYYY-MM-DD)",
    prompt="Due date (YYYY-MM-DD)",
    callback=valid_date,
    default="",
)
@click.option("--active/--inactive", "-a/-i", default=True)
def edit_target(label, description, due_date, active):
    """
    Edit the given target
    """
    repo_info = find_repo_helper(error_on_no_identity=True)
    vals = {"description": description, "due_date": due_date, "active": active}
    try:
        commands.edit_target(
            repo_info["root"], label, {k: v for k, v in vals.items() if v != ""}
        )
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)
    click.echo(colored("Updated target %s" % target, "green"))


@click.command()
def version():
    """
    Print version number and exit
    """
    print("hive version", __version__)


@click.command()
@click.option(
    "-m", "--message", help="The commit message to use", type=str, default=None
)
@click.option(
    "-a", "--add-only", help="Don't actually commit; just add files", is_flag=True
)
@click.option(
    "--pack/--no-pack",
    help="Flag to decide whether to pack untracked files or not (default behavior is to pack them together before committing)",
    default=True,
)
@click.option(
    "--pack-all",
    help="Pack all actions together (including those already tracked by the VCS) instead of just packing untracked files.  May lead to merge conflicts!!!",
    is_flag=True,
)
def commit(message, add_only, pack, pack_all):
    """
    Add all Hive-related files to the repository and commit.
    """
    repo_info = find_repo_helper(error_on_no_identity=True)

    if pack:
        vcs.pack(repo_info, untracked_only=(not pack_all))

    vcs.add(repo_info)
    if not add_only:
        vcs.commit(repo_info, message)


@click.command(name="build-cache")
def build_cache():
    """
    Cache the state of issues in the tracker
    """
    repo_info = find_repo_helper(error_on_no_identity=False)
    data.cached_all_issues(repo_info["root"], override_cache=True, progress=True)


@click.command()
@click.argument("issue")
@click.argument("fname", nargs=-1)
def pack(issue, fname):
    """
    Combine actions
    """
    repo_info = find_repo_helper(error_on_no_identity=True)

    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    try:
        new_id, num_actions = data.pack(
            repo_info["root"], issue["id"], "all" if not fname else fname
        )
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    click.echo(colored("Packed %d actions into %s" % (num_actions, new_id), "green"))


@click.command()
@click.argument("issue")
@click.argument("fname")
def unpack(issue, fname):
    """
    Combine actions
    """
    repo_info = find_repo_helper(error_on_no_identity=True)

    try:
        issue = data.get_issue_state(repo_info["root"], issue)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    try:
        new_ids, num_actions = data.unpack(repo_info["root"], issue["id"], fname)
    except data.HiveDataException as e:
        click.echo(colored("ERROR: %s" % e.args[0], "red"), err=True)
        sys.exit(1)

    click.echo(
        colored("Unpacked %s into the following action files:" % num_actions, "green")
    )
    for n in new_ids:
        click.echo(colored(i, "green"))


main.add_command(init)
main.add_command(new_issue)
main.add_command(show)
main.add_command(list_)
main.add_command(comment)
main.add_command(label)
main.add_command(close)
main.add_command(open_)
main.add_command(priority)
main.add_command(target)
main.add_command(assign)
main.add_command(web)
main.add_command(edit_label)
main.add_command(edit_target)
main.add_command(version)
main.add_command(commit)
main.add_command(build_cache)
main.add_command(pack)
main.add_command(unpack)



import re
import cgi
import shlex
import markdown

from datetime import datetime
from flask import render_template, request
from markdown.extensions import tables, fenced_code, sane_lists

from . import app
from .. import __version__ as version
from .. import data
from .. import commands
from .. import util


def do_markdown(x):
    start = markdown.markdown(
        x,
        extensions=[
            tables.TableExtension(),
            fenced_code.FencedCodeExtension(),
            sane_lists.SaneListExtension(),
        ],
    )
    return start.replace("<script", "&lt;script").replace("</script", "&lt;/script")


@app.route("/")
def list_issues():
    read_only = app.config.get("read_only", False)
    repo_info = app.config["repo_info"]
    all_labels = data.label_information(repo_info["root"])
    sorts = [i for i in request.args.get("sorts", "").split(",") if i]
    filters = shlex.split(request.args.get("filters", ""))
    new_filters = []
    for i in filters:
        i = i.lower()
        for filt in data.filter_funcs:
            if i.startswith("%s:" % filt):
                new_filters.append(tuple(i.split(":", 1)))
                break
        else:
            new_filters.append(("any", i))
    all_issues = data.all_issues_state(
        repo_info["root"], sorts=sorts, filters=new_filters
    )
    prefixes = util.unique_prefixes(data.list_issues(repo_info["root"]))
    for i in all_issues:
        i["shortid"] = prefixes[i["id"]]

        i["time"] = util.format_time(i["updated_times"][0][1])
    return render_template(
        "main.html",
        repo_info=repo_info,
        version=version,
        issues=all_issues,
        filters=request.args.get("filters", ""),
        markdown=do_markdown,
        len=len,
        sorted=sorted,
        read_only=read_only,
        all_labels=all_labels,
        url_root=app.config.get("url_root", ""),
    )


def render_comment(issue_id, c, readonly):
    return """<div class="comment" id="comment-%s"><p><small>%s commented at %s:</small></p>
    %s
    %s
    <p><small><a href="/issue/%s#comment-%s">Permalink</a></small></p>
    </div>""" % (
        c["id"],
        cgi.escape(c["authors"][0]),
        util.format_time(c["display_time"]),
        do_markdown(c["body"]),
        "\n".join(
            render_comment(issue_id, child, readonly) for child in c.get("children", [])
        ),
        issue_id,
        c["id"],
    )


@app.route("/issue/<issue_id>")
def show_issue(issue_id):
    read_only = app.config.get("read_only", False)
    repo_info = app.config["repo_info"]
    all_labels = data.label_information(repo_info["root"])
    issue = data.get_issue_state(repo_info["root"], issue_id)
    issue["time"] = util.format_time(issue["updated_times"][0][1])
    issue["utime"] = util.format_time(issue["updated_times"][-1][1])
    issue["rendered_comments"] = "\n\n".join(
        render_comment(issue["id"], i, read_only) for i in issue["comments"]
    )
    return render_template(
        "issue.html",
        repo_info=repo_info,
        version=version,
        issue=issue,
        markdown=do_markdown,
        len=len,
        read_only=read_only,
        sorted=sorted,
        all_labels=all_labels,
        url_root=app.config.get("url_root", ""),
    )



# This file is part of hive, a distributed bug tracker
# Copyright (c) 2019 by Adam Hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Python implementations of interactions with hive repositories.
"""

import hashlib

from . import data
from . import vcs

initialize = data.initialize


def new_issue(repo_root, author, title, body, parameters):
    parameters.update({"title": title, "body": body, "author": author})

    suffix = hashlib.md5(data.json_format(parameters).encode("utf-8")).hexdigest()
    new_id = None
    while new_id is None or new_id in data.list_issues(repo_root):
        new_id = data.new_issue_id(suffix)

    data.create_issue(repo_root, new_id)

    act = {"action": "set_parameters", "parameters": parameters}
    return new_id, data.add_action(repo_root, new_id, act)


def close_issue(repo_root, author, issue_id):
    return data.add_action(repo_root, issue_id, {"action": "close", "author": author})


def open_issue(repo_root, author, issue_id):
    return data.add_action(repo_root, issue_id, {"action": "open", "author": author})


def add_labels(repo_root, author, issue_id, labels):
    return data.add_action(
        repo_root, issue_id, {"action": "add_labels", "labels": set(labels)}
    )


def remove_labels(repo_root, author, issue_id, labels):
    return data.add_action(
        repo_root, issue_id, {"action": "remove_labels", "labels": set(labels)}
    )


def add_assignees(repo_root, author, issue_id, assignees):
    return data.add_action(
        repo_root, issue_id, {"action": "add_assignees", "assignees": set(assignees)}
    )


def remove_assignees(repo_root, author, issue_id, assignees):
    return data.add_action(
        repo_root, issue_id, {"action": "remove_assignees", "assignees": set(assignees)}
    )


def set_target(repo_root, author, issue_id, target):
    return data.add_action(
        repo_root,
        issue_id,
        {
            "action": "set_parameters",
            "author": author,
            "parameters": {"target": target},
        },
    )


def set_priority(repo_root, author, issue_id, priority):
    return data.add_action(
        repo_root,
        issue_id,
        {
            "action": "set_parameters",
            "author": author,
            "parameters": {"priority": priority},
        },
    )


def add_comment(repo_root, author, issue_id, body, parent=None):
    return data.add_action(
        repo_root,
        issue_id,
        {"action": "add_comment", "author": author, "body": body, "parent": parent},
    )


def edit_comment(repo_root, author, issue_id, comment_id, body):
    return data.add_action(
        repo_root,
        issue_id,
        {"action": "edit_comment", "author": author, "id": comment_id, "body": body},
    )


edit_label = data.edit_label
edit_target = data.edit_target



# This file is part of hive, a distributed bug tracker
# Copyright (c) 2019 by Adam Hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import subprocess

from setuptools import setup

from hive_bt import __version__ as HIVE_VERSION


def main():
    with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), "r") as f:
        requirements = f.read().split("\n")

    with open(os.path.join(os.path.dirname(__file__), "README"), "r") as f:
        readme = f.read()

    setup(
        name="hive_bt",
        version=HIVE_VERSION,
        author="Adam Hartz",
        author_email="hz@mit.edu",
        packages=["hive_bt", "hive_bt.web"],
        scripts=[],
        url="https://hz.mit.edu/hive",
        license="AGPLv3+",
        description="Hive is a distributed bug tracker.",
        long_description=readme,
        include_package_data=True,
        entry_points={"console_scripts": ["hive = hive_bt.cli:main"]},
        install_requires=requirements,
        package_dir={"hive_bt": "hive_bt"},
        classifiers=[
            "Development Status :: 2 - Pre-Alpha",
            "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
            "Environment :: Console",
            "Environment :: Web Environment",
            "Intended Audience :: Developers",
            "Programming Language :: Python :: 3 :: Only",
            "Topic :: Software Development :: Bug Tracking",
        ],
    )


if __name__ == "__main__":
    main()



# This file is part of hive, a distributed bug tracker
# Copyright (c) 2019 by Adam Hartz <hz@mit.edu>
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Utilities that don't belong anywhere else.
"""

import os
import shutil
import tempfile
import textwrap
import subprocess

from datetime import datetime


def find_editor():
    if "EDITOR" in os.environ:
        return shlex.split(os.environ["EDITOR"])

    for cmd in ("editor", "vim", "emacs", "nano", "vi"):
        ed = shutil.which(cmd)
        if ed:
            return [ed]

    return None


def get_editor_markdown(initial_md="", quoted_md=""):
    editor = find_editor()
    if editor is None:
        raise FileNotFoundError("could not find a valid editor.  Please set $EDITOR.")

    with tempfile.NamedTemporaryFile(mode="r+", suffix=".md") as f:
        print(initial_md, file=f)
        print(file=f)
        print("--", file=f)
        print(
            textwrap.fill(
                quoted_md.replace("--", ""), initial_indent="> ", subsequent_indent="> "
            )
        )
        print(
            "Enter Markdown above the `--`.  Text below the last `--` will be ignored.",
            file=f,
        )
        f.flush()

        subprocess.check_call(editor + [f.name])

        f.seek(0)
        body = f.read()

    return body


def make_trie(word_list):
    top = {"value": None, "children": {}}
    for word in word_list:
        node = top
        for char in word:
            if char not in node["children"]:
                node["children"][char] = {"value": 1, "children": {}}
            else:
                node["children"][char]["value"] += 1
            node = node["children"][char]
    return top


def unique_prefixes(word_list):
    trie = make_trie(word_list)
    out = {}
    for word in word_list:
        node = trie
        prefix = ""
        ix = 0
        while node["value"] != 1:
            prefix += word[ix]
            node = node["children"][word[ix]]
            ix += 1
        out[word] = prefix
    return out


def format_time(t):
    return datetime.fromtimestamp(t).strftime("%d %b %Y; %I:%M %p")


def color_256(hex_):
    r = int(hex_[:2], 16)
    g = int(hex_[2:4], 16)
    b = int(hex_[4:6], 16)
    return 16 + round(r / 51) * 36 + round(g / 51) * 6 + round(b / 51)
